package ru.dbotthepony.kstarbound.server.network

import it.unimi.dsi.fastutil.longs.LongOpenHashSet
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import ru.dbotthepony.kstarbound.client.network.packets.ForgetChunkPacket
import ru.dbotthepony.kstarbound.client.network.packets.InitialChunkDataPacket
import ru.dbotthepony.kstarbound.network.Player
import ru.dbotthepony.kstarbound.server.StarboundServer
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kvector.vector.Vector2d
import kotlin.concurrent.withLock

class ServerPlayer(connection: ServerConnection) : Player<ServerConnection>(connection, connection.uuid) {
	inline val server: StarboundServer
		get() = connection.server

	var world: ServerWorld? = null
		set(value) {
			field = value
			needsToRecomputeTrackedChunks = true
		}

	var trackedPosition: Vector2d = Vector2d(238.0, 685.0)
		set(value) {
			if (field != value) {
				field = value
				needsToRecomputeTrackedChunks = true
			}
		}

	var trackedPositionChunk: ChunkPos = ChunkPos.ZERO
		private set

	var trackedChunksWidth = 1
		set(value) {
			if (field != value) {
				field = value
				needsToRecomputeTrackedChunks = true
			}
		}

	var trackedChunksHeight = 1
		set(value) {
			if (field != value) {
				field = value
				needsToRecomputeTrackedChunks = true
			}
		}

	private val tickets = Object2ObjectOpenHashMap<ChunkPos, ServerWorld.ITicket>()
	private val sentChunks = ObjectOpenHashSet<ChunkPos>()

	private var needsToRecomputeTrackedChunks = true

	private fun recomputeTrackedChunks() {
		val world = world ?: return
		val trackedPositionChunk = world.geometry.chunkFromCell(trackedPosition)
		needsToRecomputeTrackedChunks = false
		if (trackedPositionChunk == this.trackedPositionChunk) return

		val tracked = ObjectOpenHashSet<ChunkPos>()

		for (x in -trackedChunksWidth .. trackedChunksWidth) {
			for (y in -trackedChunksHeight .. trackedChunksHeight) {
				tracked.add(world.geometry.wrap(trackedPositionChunk + ChunkPos(x, y)))
			}
		}

		val itr = tickets.entries.iterator()

		for ((pos, ticket) in itr) {
			if (pos !in tracked) {
				ticket.cancel()
				itr.remove()
			}
		}

		for (pos in tracked) {
			if (pos !in tickets) {
				tickets[pos] = world.permanentChunkTicket(pos)
			}
		}
	}

	fun tick() {
		val world = world

		if (world == null) {
			tickets.values.forEach { it.cancel() }
			tickets.clear()
			sentChunks.clear()
			return
		}

		if (needsToRecomputeTrackedChunks) {
			recomputeTrackedChunks()
		}

		for (pos in tickets.keys) {
			if (pos !in sentChunks) {
				val chunk = world.chunkMap[pos]

				if (chunk != null) {
					connection.send(InitialChunkDataPacket(chunk))
					sentChunks.add(pos)
				}
			}
		}

		val itr = sentChunks.iterator()

		for (pos in itr) {
			if (pos !in tickets) {
				connection.send(ForgetChunkPacket(pos))
				itr.remove()
			}
		}
	}
}
